// monitor.c -- Defines functions for writing to the monitor.
//             heavily based on Bran's kernel development tutorials,
//             but rewritten for JamesM's kernel tutorials.

#include "Monitor.h"

// Initializes monitor using the given details.
void monitor_init(int backcolor, int forecolor)
{
   // The VGA framebuffer starts at 0xB8000.
   videomem = (u16int *)0xB8000;
   // Stores the cursor position.
   cursor_x = 0;
   cursor_y = 0;
   
   attributeByte = (backcolor << 4) | (forecolor & 0x0F);
   attribute = attributeByte << 8;

   monitor_clear();
}


// Change the attribute settings.
void monitor_attrib(int backcolor, int forecolor)
{
   attributeByte = (backcolor << 4) | (forecolor & 0x0F);
   attribute = attributeByte << 8;
}

// Updates the hardware cursor.
static void move_cursor()
{
    // The screen is 80 characters wide...
    u16int cursorLocation = cursor_y * 80 + cursor_x;
    outb(0x3D4, 14);                  // Tell the VGA board we are setting the high cursor byte.
    outb(0x3D5, cursorLocation >> 8); // Send the high cursor byte.
    outb(0x3D4, 15);                  // Tell the VGA board we are setting the low cursor byte.
    outb(0x3D5, cursorLocation);      // Send the low cursor byte.
}

// Scrolls the text on the screen up by one line.
static void scroll()
{

    // Get a space character with the default colour attributes.
    u16int blank = 0x20 /* space */ | attribute;

    // Row 25 is the end, this means we need to scroll up
    if(cursor_y >= 25)
    {
        // Move the current text chunk that makes up the screen
        // back in the buffer by a line
        int i;
        for (i = 0*80; i < 24*80; i++)
        {
            videomem[i] = videomem[i+80];
        }

        // The last line should now be blank. Do this by writing
        // 80 spaces to it.
        for (i = 24*80; i < 25*80; i++)
        {
            videomem[i] = blank;
        }
        // The cursor should now be on the last line.
        cursor_y = 24;
    }
}

// Writes a single character out to the screen.
void monitor_put(char c)
{
    u16int *location;

    // Handle a backspace, by moving the cursor back one space
    if (c == 0x08 && cursor_x)
    {
        cursor_x--;
    }

    // Handle a tab by increasing the cursor's X, but only to a point
    // where it is divisible by 8.
    else if (c == 0x09)
    {
        cursor_x = (cursor_x+8) & ~(8-1);
    }

    // Handle carriage return
    else if (c == '\r')
    {
        cursor_x = 0;
    }

    // Handle newline by moving cursor back to left and increasing the row
    else if (c == '\n')
    {
        cursor_x = 0;
        cursor_y++;
    }
    // Handle any other printable character.
    else if(c >= ' ')
    {
        location = videomem + (cursor_y*80 + cursor_x);
        *location = c | attribute;
        cursor_x++;
    }

    // Check if we need to insert a new line because we have reached the end
    // of the screen.
    if (cursor_x >= 80)
    {
        cursor_x = 0;
        cursor_y ++;
    }

    // Scroll the screen if needed.
    scroll();
    // Move the hardware cursor.
    move_cursor();

}

// Clears the screen, by copying lots of spaces to the framebuffer.
void monitor_clear()
{
    // Make an attribute byte for the default colours
    u16int blank = 0x20 /* space */ | attribute;

    int i;
    for (i = 0; i < 80*25; i++)
    {
        videomem[i] = blank;
    }

    // Move the hardware cursor back to the start.
    cursor_x = 0;
    cursor_y = 0;
    move_cursor();
}

// Outputs a null-terminated ASCII string to the monitor.
void monitor_write(char *c)
{
    int i = 0;
    while (c[i])
    {
        monitor_put(c[i++]);
    }
}

// Output a null-terminated ASCII string to the monitor, with special attributes, applied on the whole line.
void monitor_write_on_line(char *c, int line, int backcolor, int forecolor)
{
   int i;

   u16int tmp_attrib = attribute;
   u8int tmp_attribbyte = attributeByte;

   cursor_y = line;
   cursor_x = 0;

   attributeByte = (backcolor << 4) | (forecolor & 0x0F);
   attribute = attributeByte << 8;
   
   monitor_write(c);

   for(i = cursor_x; i < 80; i++)
     monitor_put(' ');

   cursor_y = line + 1;
   cursor_x = 0;

   attributeByte = tmp_attribbyte;
   attribute = tmp_attrib; 

   move_cursor();
}

// Write a integer as hexadecimal to the screen.
void monitor_write_hex(u32int n)
{
    s32int tmp;

    monitor_write("0x"); // Writes Hex prefix to the screen.

    char noZeroes = 1;

    int i;
    for (i = 28; i > 0; i -= 4)
    {
        tmp = (n >> i) & 0xF;
        if (tmp == 0 && noZeroes != 0)
        {
            continue;
        }
    
        if (tmp >= 0xA)
        {
            noZeroes = 0;
            monitor_put (tmp-0xA+'a' );
        }
        else
        {
            noZeroes = 0;
            monitor_put( tmp+'0' );
        }
    }
  
    tmp = n & 0xF;
    if (tmp >= 0xA)
    {
        monitor_put (tmp-0xA+'a');
    }
    else
    {
        monitor_put (tmp+'0');
    }
}

// Write a integer as decimal to the screen.
void monitor_write_dec(u32int n)
{
    if (n == 0)
    {
        monitor_put('0');
        return;
    }

    s32int acc = n;
    char c[32];
    int i = 0;
    while (acc > 0)
    {
        c[i] = '0' + acc%10; 	// Puts inverse number in *c
        acc /= 10; 	
        i++;
    }
    c[i] = 0;

    char c2[32];
    c2[i--] = 0;
    int j = 0;
    while(i >= 0)
    {
        c2[i--] = c[j++];  	// Inverses the number to make it fine.
    }
    monitor_write(c2);
}
